from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestBulkClassification(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_run = cls.worker_version.worker_runs.get()
        cls.dog_class = cls.corpus.ml_classes.create(name='dog')
        cls.cat_class = cls.corpus.ml_classes.create(name='cat')

    def test_requires_login(self):
        response = self.client.post(reverse('api:classification-bulk'), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_wrong_acl(self):
        self.client.force_login(self.user)
        private_page = self.private_corpus.elements.create(
            type=self.private_corpus.types.create(slug='page'),
        )
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(private_page.id),
                    'worker_version': str(self.worker_version.id),
                    'classifications': [
                        {
                            'ml_class': str(self.dog_class.id),
                            'confidence': 0.99,
                        },
                    ],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(
            response.json(),
            {
                'parent': ['Invalid pk "{}" - object does not exist.'.format(private_page.id)]
            }
        )

    def test_worker_version(self):
        """
        Classifications are created and linked to a worker version
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(self.page.id),
                    'worker_version': str(self.worker_version.id),
                    'classifications': [
                        {
                            'ml_class': str(self.dog_class.id),
                            "confidence": 0.99,
                            "high_confidence": True
                        },
                        {
                            'ml_class': str(self.cat_class.id),
                            "confidence": 0.42,
                        }
                    ],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_cl, second_cl = self.page.classifications.order_by('-confidence').values_list('id', flat=True)
        self.assertEqual(response.json(), {
            'parent': str(self.page.id),
            'worker_version': str(self.worker_version.id),
            'worker_run_id': None,
            'classifications': [
                {
                    'id': str(first_cl),
                    'ml_class': str(self.dog_class.id),
                    'confidence': 0.99,
                    'high_confidence': True,
                    'state': 'pending',
                },
                {
                    'id': str(second_cl),
                    'ml_class': str(self.cat_class.id),
                    'confidence': 0.42,
                    'high_confidence': False,
                    'state': 'pending',
                },
            ]
        })

        self.assertCountEqual(
            list(self.page.classifications.values_list(
                'ml_class__name',
                'confidence',
                'high_confidence',
                'worker_version'
            )),
            [
                ('dog', 0.99, True, self.worker_version.id),
                ('cat', 0.42, False, self.worker_version.id)
            ],
        )

    def test_worker_version_xor_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            'ml_class': str(self.cat_class.id),
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": str(self.worker_run.id),
                    "worker_version": str(self.worker_version.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'non_field_errors': ['Only one of `worker_version` and `worker_run_id` may be set.'],
        })

    def test_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            'ml_class': str(self.dog_class.id),
                            "confidence": 0.99,
                            "high_confidence": True
                        },
                        {
                            'ml_class': str(self.cat_class.id),
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": str(self.worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_cl, second_cl = self.page.classifications.order_by('-confidence').values_list('id', flat=True)
        self.assertEqual(response.json(), {
            'parent': str(self.page.id),
            'worker_version': None,
            'worker_run_id': str(self.worker_run.id),
            'classifications': [
                {
                    'id': str(first_cl),
                    'ml_class': str(self.dog_class.id),
                    'confidence': 0.99,
                    'high_confidence': True,
                    'state': 'pending',
                },
                {
                    'id': str(second_cl),
                    'ml_class': str(self.cat_class.id),
                    'confidence': 0.42,
                    'high_confidence': False,
                    'state': 'pending',
                },
            ]
        })

        self.assertCountEqual(
            list(self.page.classifications.values_list(
                'ml_class__name',
                'confidence',
                'high_confidence',
                'worker_version_id',
                'worker_run_id',
            )),
            [
                ('dog', 0.99, True, self.worker_version.id, self.worker_run.id),
                ('cat', 0.42, False, self.worker_version.id, self.worker_run.id),
            ],
        )

    def test_worker_run_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            'ml_class': str(self.dog_class.id),
                            "confidence": 0.99,
                            "high_confidence": True
                        },
                        {
                            'ml_class': str(self.cat_class.id),
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.'],
        })

    def test_ml_class_not_found(self):
        self.dog_class.delete()
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(self.page.id),
                    'worker_version': str(self.worker_version.id),
                    'classifications': [
                        {
                            'ml_class': "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                            'confidence': 0.99,
                        },
                    ],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'classifications': ['Some ML classes do not exist or are not linked to this corpus.'],
        })

    def test_delete_worker_version(self):
        """
        Test the bulk classification API deletes previous classifications with a similar worker version
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(self.page.id),
                    'worker_version': str(self.worker_version.id),
                    'classifications': [
                        {
                            'ml_class': str(self.dog_class.id),
                            "confidence": 0.99,
                            "high_confidence": True,
                        },
                        {
                            'ml_class': str(self.cat_class.id),
                            'confidence': 0.42,
                        },
                    ],
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        another_cat_class = self.corpus.ml_classes.create(name='calico')
        best_cat_class = self.corpus.ml_classes.create(name='callico')

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(self.page.id),
                    'worker_version': str(self.worker_version.id),
                    'classifications': [
                        {
                            'ml_class': str(another_cat_class.id),
                            "confidence": 0.5,
                        },
                        {
                            'ml_class': str(best_cat_class.id),
                            "confidence": 0.85,
                            "high_confidence": True,
                        },
                    ],
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertCountEqual(
            list(self.page.classifications.values_list('ml_class__name', 'confidence', 'high_confidence')),
            [
                ('calico', 0.5, False),
                ('callico', 0.85, True),
            ],
        )

    def test_delete_worker_run(self):
        """
        Test the bulk classification API deletes previous classifications with the same worker run
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            'ml_class': str(self.dog_class.id),
                            "confidence": 0.99,
                            "high_confidence": True
                        },
                        {
                            'ml_class': str(self.cat_class.id),
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": str(self.worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        another_cat_class = self.corpus.ml_classes.create(name='calico')
        best_cat_class = self.corpus.ml_classes.create(name='callico')

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(self.page.id),
                    'classifications': [
                        {
                            'ml_class': str(another_cat_class.id),
                            "confidence": 0.5,
                        },
                        {
                            'ml_class': str(best_cat_class.id),
                            "confidence": 0.85,
                            "high_confidence": True,
                        },
                    ],
                    'worker_run_id': str(self.worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertCountEqual(
            list(self.page.classifications.values_list('ml_class__name', 'confidence', 'high_confidence')),
            [
                ('calico', 0.5, False),
                ('callico', 0.85, True),
            ],
        )

    def test_no_duplicates(self):
        """
        Test the bulk classification API prevents creating classifications with duplicate ML classes
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(self.page.id),
                    'worker_version': str(self.worker_version.id),
                    'classifications': [
                        {
                            'ml_class': str(self.dog_class.id),
                            'confidence': 0.99,
                        },
                        {
                            'ml_class': str(self.dog_class.id),
                            'confidence': 0.99,
                        },
                    ]
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'classifications': ['Duplicated ML classes are not allowed from the same worker version or worker run.']
        })
