import logging

from django.db.models.signals import pre_save
from django.dispatch import receiver
from rest_framework.exceptions import ValidationError

from arkindex.process.models import ActivityState, Process, WorkerActivityState, WorkerConfiguration, WorkerRun
from arkindex.process.utils import hash_object
from ponos.signals import task_failure

logger = logging.getLogger(__name__)


def _list_ancestors(graph, parents):
    results = []

    for parent_id in parents:
        results.append(parent_id)
        results.extend(_list_ancestors(graph, graph[parent_id]))

    return results


@receiver(pre_save, sender=WorkerRun)
def check_parents(sender, instance, **kwargs):
    """
    Since we can't properly validate parents field on WorkerRun by simply defining an Array of UUIDs,
    we add a pre-save signal with constraints.
    It will check that provided UUIDs refer to existing WorkerRun objects and that these objects are
    linked to the same Process as the to-be-saved WorkerRun instance.
    It will also check that adding or updating WorkerRun parents will not create cycles in the tree.
    """
    if not instance.parents:
        return

    # Use `process_id` and not `process` to avoid an extra SQL query that would use the replica
    parents = WorkerRun.objects.using('default') \
        .filter(id__in=instance.parents, process_id=instance.process_id) \
        .exclude(id=instance.id).values_list('id', flat=True)

    if set(parents) != set(instance.parents):
        raise ValidationError(f"Can't add or update WorkerRun {instance.id} because parents field isn't properly defined. It can be either because"
                              " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
                              " same Process than this WorkerRun.")

    graph = {
        wr['id']: wr['parents']
        for wr in WorkerRun.objects.using('default').filter(process_id=instance.process_id).values('id', 'parents')
    }
    ancestors = _list_ancestors(graph, instance.parents)
    if instance.id in ancestors:
        raise ValidationError(f"Can't add or update WorkerRun {instance.id} because parents field isn't properly defined. It would create a cycle.")


@receiver(pre_save, sender=WorkerConfiguration)
def update_configuration_hash(sender, instance, **kwargs):
    instance.configuration_hash = hash_object(instance.configuration)


@receiver(pre_save, sender=WorkerRun)
def generate_summary(sender, instance, **kwargs):
    # This avoids generating summaries when loading fixtures
    if kwargs['raw']:
        return
    instance.build_summary()


@receiver(task_failure)
def stop_started_activities(sender, task, **kwargs):
    """
    When a Ponos task fails, update any relevant WorkerActivity instances from `started` to `error`.
    This allows to retry a process and get it to re-run on activities that were not finished without skipping.

    This is definitely not perfect, as this can cause WorkerActivities for worker versions of unrelated tasks
    to also be marked as error when they are successful, but we do not have any reliable link
    between Ponos tasks and Arkindex worker versions.
    """
    process = Process.objects.filter(workflow_id=task.workflow_id).only('id', 'activity_state').first()
    if not process or process.activity_state == ActivityState.Disabled:
        return
    count = process.activities.filter(state=WorkerActivityState.Started).update(state=WorkerActivityState.Error)
    logger.info(f'Updated {count} worker activities from Started to Error as a task failed on {process.id}')
